/* Copyright ORNL */

#ifndef MINIAPPS_ALGORITHMS_TRICOUNT_HPP_
#define MINIAPPS_ALGORITHMS_TRICOUNT_HPP_

#include <slepceps.h>
#include <petscmat.h>
#include <armadillo>
#include "mpicomm.hpp"
#include "utils.h"


// #define TRICOUNT_VERBOSE 1

/*
 * 1D distribution
 * number of processors p
 * n number of nodes in the graph
 * Inputs is a 1D distributed adjacency matrix for a 'n' vertex graph
 * A is an incidence matrix.
 * A is n/p * n
 * All 'p' processors will have same vector
 */

class TriangleCount {
 private:
  const MPICommunicator& m_mpicomm;

  Mat A;   /* operator matrix */
  EPS eps; /* eigenproblem solver context */
  EPSType   type;
  arma::sp_fmat At;  // May not be needed in other non-arma cases.

  PetscInt local_m, local_n;
  PetscInt global_m, global_n;
  PetscBool terse;
  PetscErrorCode ierr;
  double m_num_triangles = 0;

  void debug_petsc_mat() {
    // understanding matrix properties.
    MatGetSize(A, &global_m, &global_n);
    MatGetLocalSize(A, &local_m, &local_n);
    PRINTROOT("global m::" << this->global_m << "::n::" << global_n);
    PRINTROOT("local m::" << this->local_m << "::n::" << local_n);

    MatGetOwnershipRange(A, &local_m, &local_n);
    DISTPRINTINFO("matrange::" << local_m << "," << local_n);
    PetscPrintf(PETSC_COMM_WORLD, "Matrix created\n");
#ifdef TRICOUNT_VERBOSE
    MatView(A, PETSC_VIEWER_STDOUT_WORLD);
#endif
    //quality of petsc initialization
    Vec v0, v1;
    ierr = MatCreateVecs(A, &v0, &v1);  // CHKERRQ(ierr);
    ierr = VecSet(v0, 1);  // CHKERRQ(ierr);
    MatMult(A, v0, v1);

    PetscInt init_vec_size;
    VecGetSize(v0, &init_vec_size);
    PetscPrintf(PETSC_COMM_WORLD, "init vector created\n");
    PRINTROOT("initsize::" << init_vec_size);
    VecGetOwnershipRange(v1, &local_m, &local_n);
    DISTPRINTINFO("vecrange::" << local_m << "," << local_n);
#ifdef TRICOUNT_VERBOSE
    VecView(v0, PETSC_VIEWER_STDOUT_WORLD);
    VecView(v1, PETSC_VIEWER_STDOUT_WORLD);
    Mat local_mat;
    MatCreate(PETSC_COMM_SELF, &local_mat);
    MatSetType(local_mat, MATSEQAIJ);

    MatMPIAIJGetLocalMat(A, MAT_INITIAL_MATRIX, &local_mat);
    for (int i = 0; i < MPI_SIZE; i++) {
      if (MPI_RANK == i) {
        MatView(local_mat, PETSC_VIEWER_STDOUT_SELF);
      }
      MPI_Barrier(MPI_COMM_WORLD);
    }
    ierr = MatDestroy(&local_mat);
#endif
    ierr = VecDestroy(&v0);
    ierr = VecDestroy(&v1);
  }

 public:
  TriangleCount(int n_rows, int n_cols, int n_nnz,
                PetscInt *row_ptrs, PetscInt *col_idx, PetscScalar *val,
                PetscInt *nnz_per_row, const MPICommunicator& communicator):
    m_mpicomm(communicator) {
    int global_rows = 0;
    MPI_Allreduce(&n_rows, &global_rows, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
    MatCreate(PETSC_COMM_WORLD, &A);
    MatSetType(A, MATMPIAIJ);
    MatSetSizes(A, n_rows, PETSC_DECIDE, global_rows, n_cols);
    MatMPIAIJSetPreallocationCSR(A, row_ptrs, col_idx, val);
#ifdef TRICOUNT_VERBOSE
    for (int i = 0; i < n_rows + 1; i++) {
      cout << MPI_RANK << "::" << "row_ptrs::" << row_ptrs[i] << endl;

    }
    for (int i = 0; i < n_nnz; i++) {
      cout << MPI_RANK << "::" << "val::" << val[i] << endl;
      cout << MPI_RANK << "::" << "col_idx::" << col_idx[i] << endl;
    }
#endif
    ierr = MatCreateMPIAIJWithArrays(PETSC_COMM_WORLD, n_rows, PETSC_DECIDE,
                                     n_rows * communicator.size(), n_cols,
                                     row_ptrs, col_idx, val,
                                     &A);
#ifdef TRICOUNT_VERBOSE
    debug_petsc_mat();
#endif
  }

  TriangleCount(int n_rows, int n_cols, int n_nnz,
                PetscInt *row_idx, PetscInt *col_idx, PetscScalar *val,
                const MPICommunicator& communicator):
    m_mpicomm(communicator) {
    // call in the driver program itself
    // SlepcInitialize(&argc, &argv, (char *)0, help);
    // MatMPIAIJSetPreallocationCSR(this->A, row_ptrs, col_idx, val);

    //compute matrix statistics. Use for matsetvalues
    int *start_row = new int[communicator.size()];
    int *all_proc_rows = new int[communicator.size()];
    int global_rows = 0;
    MPI_Allgather(&n_rows, 1, MPI_INT, all_proc_rows, 1, MPI_INT, MPI_COMM_WORLD);
    start_row[0] = 0;
    for (int i = 0; i < communicator.size(); i++) {
      if (i > 0) {
        start_row[i] = start_row[i - 1] + all_proc_rows[i];
      }
      global_rows += all_proc_rows[i];
    }

    // find the max nnzs per row for preallocation.
    int *nnzs_per_rows = new int[n_rows];
    for (int i = 0; i < n_rows; i++) {
      nnzs_per_rows[i] = 0;
    }
    for (int i = 0; i < n_nnz; i++) {
      nnzs_per_rows[row_idx[i]]++;
    }
    int max_nnz_per_row = -1;
    for (int i = 0; i < n_rows; i++) {
      if (nnzs_per_rows[i] > max_nnz_per_row) {
        max_nnz_per_row = nnzs_per_rows[i];
      }
    }
    int max_cols;
    MPI_Allreduce(&n_cols, &max_cols, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
    //compute matrix statistics.
    // int *start_col = new int[communicator.size()];
    // int *all_proc_cols = new int[communicator.size()];
    // int global_cols = 0;
    // MPI_Allgather(&n_cols, 1, MPI_INT, all_proc_cols, 1, MPI_INT, MPI_COMM_WORLD);
    // start_col[0] = 0;
    // for (int i = 0; i < communicator.size(); i++) {
    //   if (i > 0) {
    //     start_col[i] = start_col[i - 1] + all_proc_cols[i];
    //   }
    //   global_cols += all_proc_cols[i];
    // }
    DISTPRINTINFO("rows,cols,nnzs, max_nnz_per_row" << n_rows  << ","
                  << n_cols << ","  << n_nnz << max_nnz_per_row);


    MatCreate(PETSC_COMM_WORLD, &A);
    MatSetType(A, MATMPIAIJ);
    MatSetSizes(A, n_rows, PETSC_DECIDE, global_rows, max_cols);
    MatMPIAIJSetPreallocation(A, max_nnz_per_row, PETSC_NULL, max_nnz_per_row, PETSC_NULL);
    // ierr = MatCreateMPIAIJWithArrays(PETSC_COMM_WORLD, n_rows, n_cols,
    //                                   n_rows * communicator.size(), n_cols,
    //                                   row_ptrs, col_idx, val,
    //                                   &A);
    // ierr = MatCreateMPIAIJ(PETSC_COMM_WORLD, n_rows, n_cols, PETSC_DEFAULT, nnz_per_row, &A);
    // ierr = MatCreateMPIAIJ(PETSC_COMM_WORLD, n_rows, n_cols,
    //                        n_rows * communicator.size(), n_cols,
    //                        0, PETSC_NULL, 0, PETSC_NULL, &A);

    // // CHKERRQ(ierr);
//#ifdef TRICOUNT_VERBOSE
    // for (int i = 0; i < n_nnz; i++) {
    //   cout << communicator.rank() << "::" << row_idx[i]
    //        << "," << col_idx[i] << "," << val[i] << endl;
    // }
//#endif

    PetscInt local_row_idx;
    PetscInt local_col_idx;
    PetscScalar local_val;

    int my_start_row = start_row[MPI_RANK];
    // int my_start_row = 0;
    // int my_start_col = start_col[communicator.rank()];
    int my_start_col = 0;
    // MatGetOwnershipRange(A, &my_start_row, &my_start_col);
    double petsc_insert_time = 0.0;
    for (int i = 0; i < n_nnz; i++) {
      local_row_idx = my_start_row + row_idx[i];
      local_col_idx = my_start_col + col_idx[i];
      local_val = val[i];
#ifdef TRICOUNT_VERBOSE
      DISTPRINTINFO(local_row_idx << "," << local_col_idx << "," << local_val);
#endif
      tic();
      ierr = MatSetValues(A, 1, &local_row_idx, 1, &local_col_idx, &local_val, INSERT_VALUES);
      petsc_insert_time += toc();
      if (i % 25000 == 0) {
        PRINTROOT("25000 time::" << petsc_insert_time);
        petsc_insert_time = 0;
      }
      CHKERRV(ierr);
    }
    PRINTROOT("100000 time::" << petsc_insert_time);
    tic();
    MatAssemblyBegin(A, MAT_FINAL_ASSEMBLY);
    MatAssemblyEnd(A, MAT_FINAL_ASSEMBLY);
    petsc_insert_time = toc();
    PRINTROOT("calling assembly to end::took::" << petsc_insert_time);
    PetscPrintf(PETSC_COMM_WORLD, "Matrix created\n");
// #ifdef TRICOUNT_VERBOSE
    debug_petsc_mat();
// #endif
  }
  PetscErrorCode count() {

    PetscInt nev;
    /*
    * Set the initial vector. This is optional, if not done the initial
    * vector is set to random values
    */


    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    *             Create the eigensolver and set various options
    *  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

    /*
     * Create eigensolver context
     */
    ierr = EPSCreate(PETSC_COMM_WORLD, &eps);  CHKERRQ(ierr);

    /*
     * Set operators. In this case, it is a standard eigenvalue problem
     */
    ierr = EPSSetOperators(eps, A, NULL);  CHKERRQ(ierr);
    ierr = EPSSetProblemType(eps, EPS_NHEP);  CHKERRQ(ierr);

    /*
     * Set solver parameters at runtime
     */
    ierr = EPSSetFromOptions(eps); CHKERRQ(ierr);

    /*
    * Set the nature of the eigen values suitable for triangle counting
    */
    ierr = EPSSetWhichEigenpairs(eps, EPS_LARGEST_REAL); CHKERRQ(ierr);
    ierr = EPSSetDimensions(eps, 100, PETSC_DEFAULT, PETSC_DEFAULT); CHKERRQ(ierr);
    // BV V;
    // EPSGetBV(eps, &V);
    // BVSetType(V, BVVECS);
    PRINTROOT("calling epssolve");

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     *                   Solve the eigensystem
     *  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

    ierr = EPSSolve(eps);  CHKERRQ(ierr);

    /*
     * Optional: Get some information from the solver and display it
     */
    ierr = EPSGetType(eps, &type);  CHKERRQ(ierr);
    ierr = PetscPrintf(PETSC_COMM_WORLD, " Solution method: %s\n\n", type);
    // CHKERRQ(ierr);
    ierr = EPSGetDimensions(eps, &nev, NULL, NULL);  CHKERRQ(ierr);
    ierr = PetscPrintf(PETSC_COMM_WORLD,
                       " Number of requested eigenvalues: %D\n",
                       nev);  CHKERRQ(ierr);
    double sum_of_eigen_cube = 0;
    double current_eigen_valr = 0;
    double current_eigen_vali = 0;
    for (int i = 0; i < nev; i++) {
      ierr = EPSGetEigenvalue(eps, i, &current_eigen_valr, &current_eigen_vali);
      CHKERRQ(ierr);
      sum_of_eigen_cube += current_eigen_valr * current_eigen_valr
                           * current_eigen_valr;
    }
    PRINTROOT("sum_of_eigen_cube :: " << sum_of_eigen_cube);
    m_num_triangles = sum_of_eigen_cube / 3;
    return ierr;
  }

  ~TriangleCount() {
    ierr = EPSDestroy(&eps);  // CHKERRQ(ierr);
    ierr = MatDestroy(&A);  // CHKERRQ(ierr);
    // CHKERRQ(ierr);
  }

  int num_triangles() {return m_num_triangles;}
};  // class TriangleCount

#endif  // MINIAPPS_ALGORITHMS_TRICOUNT_HPP_
